*****************************************************
*                                                   *
* lock_process.asm                                  *
*                                                   *
* extended hog_processor() daemon for IC            *
*                                                   *
* Dr. Richard F. Drushel                            *
* Department of Biology                             *
* Case Western Reserve University                   *
* 10900 Euclid Avenue                               *
* Cleveland, Ohio  44106-7080  USA                  *
* rfd@po.cwru.edu                                   *
*                                                   *
*****************************************************

*****************************************************************************
*                                                                           *
* version history (add updates to top of list)                              *
*                                                                           *
* 03  Fri Mar 03 09:59:00 2000  ;fixed some comments                        *
*                                                                           *
* 02  Thu Mar 02 10:26:00 2000  ;added 32-bit counter decrement code.       *
*                               ;it works!!                                 *
*                                                                           *
* 01  Wed Mar 01 08:49:00 2000  ;made _LOCK_PROCESS_FLAG a 2-bye <int>      *
*                               ;divided <long> counter into 2 <int>s:      *
*                               ;_LOCK_PROCESS_TICK_COUNT_HI and            *
*                               ;_LOCK_PROCESS_TICK_COUNT_LO.               *
*                                                                           *
* 00  Tue Feb 29 15:04:00 2000  ;initial coding                             *
*                                                                           *
*****************************************************************************

*****************************************************************************
*                                                                           *
* Theory of Operation                                                       *
*                                                                           *
* This daemon allows whatever the current IC process is to run without      *
* swapout by the process scheduler, until a preset counter of 1-kHz system  *
* timer ticks has run down.  The counter may range from 0-4294967295 msec   *
* (0-49.71 days).                                                           *
*                                                                           *
* The existing hog_processor() function merely pokes 0 into the scheduler's *
* own 1-byte timer tick counter (at address 0x0a), thereby bumping up the   *
* timeslice another 256 msec.  For users who want to guarantee that code    *
* will not be interrupted on a longer timescale (seconds to minutes), this  *
* is (1) not enough time, and (2) requires some inconvenient code to        *
* repeatedly call hog_processor() to keep bumping up the timeslice.         *
*                                                                           *
* The lock_process daemon is executed with every timer tick.  It looks at a *
* global flag variable, _LOCK_PROCESS_FLAG, to determine if the current     *
* process has been made uninterruptable.  This flag is set and cleared by   *
* user calls to lock_process() and unlock_process(), respectively.          *
*                                                                           *
* If the flag is zero, the daemon does nothing during this timer tick.      *
*                                                                           *
* If the flag is nonzero, the daemon decrements a 4-byte (IC <long> type)   *
* counter variable, _LOCK_PROCESS_TICK_COUNT, whose initial value was set   *
* in user code by calling the lock_process(float time) function.            *
*                                                                           *
* If the counter reaches 1, a 1 is poked into the process scheduler timer   *
* tick counter (at address 0x0a), _LOCK_PROCESS_FLAG is cleared, and the    *
* counter is zeroed.  This unlocks the current process, giving it one more  *
* msec of execution time; at the next timer tick, it will be swapped out by *
* the scheduler.  Note:  this means that the current process *does* have    *
* the opportunity to re-assert another process lock (either hog_processor() *
* or lock_process()) before the next timer tick...                          *
*                                                                           *
* If the counter is still greater than 1, a 0 is poked into the process     *
* scheduler timer tick counter (at address 0x0a).  This is just what        *
* hog_processor() does to keep a process from being swapped out.            *
*                                                                           *
* IMPLEMENTATION NOTE:                                                      *
*                                                                           *
* The 4-byte <long> counter is implemented as 2 adjacent 2-byte <int>       *
* variables, _LOCK_PROCESS_COUNT_HI and _LOCK_PROCESS_COUNT_LO.  This is    *
* because (1) IC and the .ICB compiler can only create user-accessible      *
* <int> variables, and (2) IC prevents pointer arithmetic to offset to the  *
* individual bytes in a variable.                                           *
*                                                                           *
*****************************************************************************

******************************************************************
*                                                                *
* .ICB COMPILER NOTE:                                            *
*                                                                *
* All the (pound)includes were removed and the code inlined for  *
* compatibility with Newton Labs' web-based .icb compiler (which *
* doesn't support it and will barf if the string (pound)include  *
* is even present in the .asm file!)                             *
*                                                                *
******************************************************************

* (pound)include <6811regs.asm>

* file of standard 6811 register declarations

*********************************************************************
* Control Registers

BASE    EQU     $1000

PORTA   EQU     $1000   ; Port A data register
RESV1   EQU     $1001   ; Reserved
PIOC    EQU     $1002   ; Parallel I/O Control register
PORTC   EQU     $1003   ; Port C latched data register
PORTB   EQU     $1004   ; Port B data register
PORTCL  EQU     $1005   ;
DDRC    EQU     $1007   ; Data Direction register for port C
PORTD   EQU     $1008   ; Port D data register
DDRD    EQU     $1009   ; Data Direction register for port D
PORTE   EQU     $100A   ; Port E data register
CFORC   EQU     $100B   ; Timer Compare Force Register
OC1M    EQU     $100C   ; Output Compare 1 Mask register
OC1D    EQU     $100D   ; Output Compare 1 Data register

* Two-Byte Registers (High,Low -- Use Load & Store Double to access)

TCNT    EQU     $100E   ; Timer Count Register
TIC1    EQU     $1010   ; Timer Input Capture register 1
TIC2    EQU     $1012   ; Timer Input Capture register 2
TIC3    EQU     $1014   ; Timer Input Capture register 3
TOC1    EQU     $1016   ; Timer Output Compare register 1
TOC2    EQU     $1018   ; Timer Output Compare register 2
TOC3    EQU     $101A   ; Timer Output Compare register 3
TOC4    EQU     $101C   ; Timer Output Compare register 4
TI4O5   EQU     $101E   ; Timer Input compare 4 or Output compare 5 register

TCTL1   EQU     $1020   ; Timer Control register 1
TCTL2   EQU     $1021   ; Timer Control register 2
TMSK1   EQU     $1022   ; main Timer interrupt Mask register 1
TFLG1   EQU     $1023   ; main Timer interrupt Flag register 1
TMSK2   EQU     $1024   ; misc Timer interrupt Mask register 2
TFLG2   EQU     $1025   ; misc Timer interrupt Flag register 2
PACTL   EQU     $1026   ; Pulse Accumulator Control register
PACNT   EQU     $1027   ; Pulse Accumulator Count register
SPCR    EQU     $1028   ; SPI Control Register
SPSR    EQU     $1029   ; SPI Status Register
SPDR    EQU     $102A   ; SPI Data Register
BAUD    EQU     $102B   ; SCI Baud Rate Control Register
SCCR1   EQU     $102C   ; SCI Control Register 1
SCCR2   EQU     $102D   ; SCI Control Register 2
SCSR    EQU     $102E   ; SCI Status Register
SCDR    EQU     $102F   ; SCI Data Register
ADCTL   EQU     $1030   ; A/D Control/status Register
ADR1    EQU     $1031   ; A/D Result Register 1
ADR2    EQU     $1032   ; A/D Result Register 2
ADR3    EQU     $1033   ; A/D Result Register 3
ADR4    EQU     $1034   ; A/D Result Register 4
BPROT   EQU     $1035   ; Block Protect register
RESV2   EQU     $1036   ; Reserved
RESV3   EQU     $1037   ; Reserved
RESV4   EQU     $1038   ; Reserved
OPTION  EQU     $1039   ; system configuration Options
COPRST  EQU     $103A   ; Arm/Reset COP timer circuitry
PPROG   EQU     $103B   ; EEPROM Programming register
HPRIO   EQU     $103C   ; Highest Priority Interrupt and misc.
INIT    EQU     $103D   ; RAM and I/O Mapping Register
TEST1   EQU     $103E   ; factory Test register
CONFIG  EQU     $103F   ; Configuration Control Register

* Interrupt Vector locations

SCIINT  EQU     $D6     ; SCI serial system
SPIINT  EQU     $D8     ; SPI serial system
PAIINT EQU      $DA     ; Pulse Accumulator Input Edge
PAOVINT EQU     $DC     ; Pulse Accumulator Overflow
TOINT   EQU     $DE     ; Timer Overflow
TOC5INT EQU     $E0     ; Timer Output Compare 5
TOC4INT EQU     $E2     ; Timer Output Compare 4
TOC3INT EQU     $E4     ; Timer Output Compare 3
TOC2INT EQU     $E6     ; Timer Output Compare 2
TOC1INT EQU     $E8     ; Timer Output Compare 1
TIC3INT EQU     $EA     ; Timer Input Capture 3
TIC2INT EQU     $EC     ; Timer Input Capture 2
TIC1INT EQU     $EE     ; Timer Input Capture 1
RTIINT  EQU     $F0     ; Real Time Interrupt
IRQINT  EQU     $F2     ; IRQ External Interrupt
XIRQINT EQU     $F4     ; XIRQ External Interrupt
SWIINT  EQU     $F6     ; Software Interrupt
BADOPINT EQU    $F8     ; Illegal Opcode Trap Interrupt
NOCOPINT EQU    $FA     ; COP Failure (Reset)
CMEINT  EQU     $FC     ; COP Clock Monitor Fail (Reset)
RESETINT EQU    $FE     ; RESET Interrupt

* end file "6811regs.asm" ***************************************************

* required statement to start code

        ORG     MAIN_START

*****************************************************************************

* declare variables

* flag for enforcing extended process locking
*    0         don't do it
*    nonzero   do it

variable__LOCK_PROCESS_FLAG:

        FDB     0

* counter for how many timer ticks remaining for current process lock
* note:  this is a <long> for IC (4 bytes)
* remember, it's Motorola byte order (hi-lo, bigendian, blecch!)

variable__LOCK_PROCESS_TICK_COUNT_HI:

        FDB     0

variable__LOCK_PROCESS_TICK_COUNT_LO:

        FDB     0

*****************************************************************************

* initialization routine 

* (1) install into System Interrupt structure
* (2) initialize subroutine variables

subroutine_initialize_module:

* (pound)include <ldxibase.asm>

* 
* File "ldxibase.asm"
*
* Fred Martin Thu Oct 10 19:49:38 1991
*
*
* The following code loads the X register with a base pointer to
* the 6811 interrupt vectors:  $FF00 if the 6811 is in normal mode,
* and $BF00 if the 6811 is in special mode.
*
* The file "6811regs.asm" must be loaded first for this to work.
*

        LDAA    HPRIO
        ANDA    #$40                    ; test SMOD bit
        BNE     *+7
        LDX     #$FF00                  ; normal mode interrupts
        BRA     *+5
        LDX     #$BF00                  ; special mode interrupts

* end file "ldxibase.asm" ***************************************************

        LDD     TOC4INT,X               ; SystemInt on TOC4
        STD     interrupt_code_exit+1
        LDD     #interrupt_code_start
        STD     TOC4INT,X

        LDD     #$0000
        STD     variable__LOCK_PROCESS_FLAG             ;disable locking
        STD     variable__LOCK_PROCESS_TICK_COUNT_HI    ;zero hiword
        STD     variable__LOCK_PROCESS_TICK_COUNT_LO    ;zero loword

        RTS

*****************************************************************************

* main process locking code

interrupt_code_start:

        LDD     variable__LOCK_PROCESS_FLAG
        CMPD    #$0000                          ;are we locking this process?
        BEQ     interrupt_code_exit             ;no, so exit immediately

* process is locked, so decrement the timer tick counter

        LDD     variable__LOCK_PROCESS_TICK_COUNT_LO    ;get loword
        SUBD    #$0001                                  ;one less
        STD     variable__LOCK_PROCESS_TICK_COUNT_LO    ;save it back
        BCC     check_counter                           ;no borrow from hiword
        LDD     variable__LOCK_PROCESS_TICK_COUNT_HI    ;get hiword
        SUBD    #$0001                                  ;one less
        STD     variable__LOCK_PROCESS_TICK_COUNT_HI    ;save it back

* have we ticked down to 1?  if so, hiword=0, loword=1.

check_counter:

        LDAA    #0                                      ;assume 256 msec left for
*                                                       ;process scheduler timer
        LDD     variable__LOCK_PROCESS_TICK_COUNT_HI    ;get hiword
        CMPD    #0000                                   ;should be zero if we're done
        BNE     not_done_yet                            ;nope
        LDD     variable__LOCK_PROCESS_TICK_COUNT_LO    ;get loword
        CMPD    #0001                                   ;is it 1?
        BNE     not_done_yet                            ;nope

* timer has expired, so remove the process lock

        CLR     variable__LOCK_PROCESS_FLAG             ;clear the lock flag hibyte
        CLR     variable__LOCK_PROCESS_FLAG+1           ;lobyte
        CLR     variable__LOCK_PROCESS_TICK_COUNT_LO+1  ;wipe the counter loword lobyte
*                                                       ;(the others are 0 already)
        INCA                                            ;0 + 1 = 1 msec left

not_done_yet:

        STAA    $000A                                   ;poke value into the scheduler timer

interrupt_code_exit:

        JMP     $0000
